// AudioDecoder provides a generic interface that uses the JavaSound API to
// automatically identify, decode, and play streamed audio data.
//
// Support for audio formats not already supported by JavaSound can typically
// be added simply by adding appropriate JavaSound SPI libraries to the
// classpath. For example, Ogg Vorbis and MP3 SPI libraries are available here:
//  http://www.javazoom.net/projects.html

package com.tightvnc.decoder.audio;

import com.tightvnc.vncviewer.RfbInputStream;
import com.tightvnc.decoder.audio.QueueInputStream;
import java.io.BufferedInputStream;
import java.io.DataOutput;
import java.io.DataInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.util.LinkedList;
import java.util.NoSuchElementException;
import javax.sound.sampled.*;

public class AudioDecoder {

  // Input / Recording Streams.
  protected RfbInputStream rfbis = null;
  protected DataInputStream dis = null;
  protected DataOutput dos = null;

  // Decoding Queue between Main and Decoding Threads.
  protected QueueInputStream decodingQueue;
  protected static int DECODING_QUEUE_LIMIT = 102400;  // in bytes

  // Decoding Thread.
  protected Thread decodingThread = null;

  // Decoded Format.
  protected AudioFormat outputFormat;

  // Output Buffer between Decoding and Output Threads.
  protected LinkedList outputBuffer;
  protected int outputBufferTargetLength = 50;  // in milliseconds
  protected int outputBufferTargetSize;  // in bytes
  protected int outputBufferSize;

  // Output Thread.
  protected Thread outputThread;
  protected SourceDataLine localSoundSystem;
  public static long statNumBytesOut = 0;  // number of bytes of decoded data written to sound card


  public AudioDecoder(RfbInputStream is) {
    setRfbInputStream(is);
    setDefaultOutputFormat();
  }

  public void setRfbInputStream(RfbInputStream is) {
    rfbis = is;
  }

  public AudioDecoder(DataInputStream is) {
    setDataInputStream(is);
    setDefaultOutputFormat();
  }

  public void setDataInputStream(DataInputStream is) {
    dis = is;
  }

  // FIXME: Rename this method after we don't need RecordInterface
  // in Decoder class to record session
  public void setDataOutputStream(DataOutput os) {
    dos = os;
  }

  //
  // Server has enabled audio streaming.
  //
  public void handleEnableMsg() {
    if (decodingThread != null) {
      System.out.println("Server attempted to enable audio, but audio is already enabled");
      return;
    }

    // Start the decoding thread, which will start the output thread.
    decodingQueue = new QueueInputStream(DECODING_QUEUE_LIMIT);
    decodingThread = new Thread(new decodingClass());
    decodingThread.start();
  }

  //
  // Server has disabled audio streaming.
  //
  public void handleDisableMsg() {
    if (decodingThread == null) {
      System.out.println("Server attempted to disable audio, but audio is already disabled");
      return;
    }

    // Stop the decoding thread, which will stop the output thread.
    decodingThread = null;
    if (decodingQueue != null) {
      try {
        decodingQueue.close();
      } catch (IOException e) {}
    }
  }

  //
  // Accept audio data from the server and push it into a queue for decoding by
  // a separate thread.  We want to minimize delays here so audio decoding and
  // output do not block other VNC operations.
  //
  public void handleMsg(int size) throws IOException, Exception {
    byte[] buf = new byte[size];
    if (rfbis != null) {
      rfbis.readFully(buf);
    } else {
      dis.readFully(buf);
    }

    // Save decoded data to record output stream.
    if (dos != null) {
      dos.write(buf);
    }

    if (decodingThread == null) {
      System.out.println("Server sent audio data, but audio is disabled");
      return;
    }

    // Drop the data if there was an audio processing error.
    if (decodingQueue == null) { return; }

    try {
      decodingQueue.write(buf);
    } catch (IOException e) {
      System.out.println("Hit audio queuing limit, dropping an audio data message");
    }
  }

  public void setDefaultOutputFormat() {
    setOutputFormat(2, 44100, 16, true, true);
  }

  public void setOutputFormat(int channels, int sampleRate, int sampleSize,
                              boolean signed, boolean bigEndian) {
    outputFormat = new AudioFormat(sampleRate, sampleSize, channels, signed, bigEndian);
    updateBufferTargetSize();
    if (decodingThread != null) {
      handleDisableMsg();
      handleEnableMsg();
    }
  }

  public void setBufferLength(int milliseconds) {
    outputBufferTargetLength = milliseconds;
    updateBufferTargetSize();
    if (decodingThread != null) {
      handleDisableMsg();
      handleEnableMsg();
    }
  }

  private void updateBufferTargetSize() {
    outputBufferTargetSize = (outputBufferTargetLength * (int)outputFormat.getSampleRate()
      * (outputFormat.getSampleSizeInBits() / 8) * outputFormat.getChannels()) / 100;
  }

  // Create a decoded InputStream in the specified format from an encoded InputStream.
  protected InputStream prepareDecoder(InputStream originalStream, AudioFormat outputFormat) {
    // Create an AudioInputStream, auto-detecting the input format.
    AudioInputStream originalAudioStream;
    try {
      originalAudioStream = AudioSystem.getAudioInputStream(new BufferedInputStream(originalStream));
    } catch (UnsupportedAudioFileException e) {
      System.out.println("Disabling audio: Server is using an unsupported audio format");
      return null;
    } catch (IOException e) {
      System.out.println("Disabling audio: IO Error while detecting audio format");
      return null;
    }

    // Create another AudioInputStream which decodes the first one.
    AudioInputStream decodedAudioStream = originalAudioStream;
    if (!outputFormat.matches(originalAudioStream.getFormat())) {
      try {
        decodedAudioStream = AudioSystem.getAudioInputStream(outputFormat, originalAudioStream);
      } catch (IllegalArgumentException e) {
        System.out.println("Disabling audio: No appropriate audio decoder is available");
        return null;
      }
    }

    return decodedAudioStream;
  }

  //
  // The audio decoding thread.
  //
  protected class decodingClass implements Runnable {
    public void run() {
      InputStream decodedAudio = prepareDecoder(decodingQueue, outputFormat);
      localSoundSystem = prepareSoundSystem(outputFormat);
      if (decodedAudio == null || localSoundSystem == null) {
        decodingQueue = null;
        return;
      }

      // Start the output thread.
      outputBufferSize = 0;
      outputBuffer = new LinkedList();
      outputThread = new Thread(new outputClass());
      outputThread.start();

      // Accept audio data from the queue, decode it, and push it into a buffer
      // for output by a separate thread.  We use our own buffer because the
      // buffering built into SourceDataLine doesn't handle streaming very well.
      byte[] readData = new byte[4096];
      int readSize = 0;
      while (true) {
        try {
          readSize = decodedAudio.read(readData);
        } catch (IOException e) {
          System.out.println("IO Error reading/decoding audio data, will retry");
          continue;
        }
        if (readSize == -1) { break; }

        byte[] data = new byte[readSize];
        System.arraycopy(readData, 0, data, 0, readSize);

        synchronized (outputBuffer) {
          while (outputBufferSize > outputBufferTargetSize &&
                 readSize + outputBufferSize > outputBufferTargetSize * 2) {
            // Buffer is double the target size, sleep until it drains some
            try { outputBuffer.wait(); } catch (InterruptedException ie) {}
          }
          outputBufferSize += readSize;
          outputBuffer.addLast(data);
          outputBuffer.notifyAll();
        }
      }

      // Stop the output thread.
      outputThread = null;
      synchronized (outputBuffer) {
        outputBuffer.notifyAll();
      }
      decodingQueue = null;
    }
  }

  // Prepare the local sound system.
  protected SourceDataLine prepareSoundSystem(AudioFormat outputFormat) {
    try {
      SourceDataLine sdl = (SourceDataLine)(AudioSystem.getLine(
        new DataLine.Info(SourceDataLine.class, outputFormat)));
      sdl.open(outputFormat);
      return sdl;
    } catch (IllegalArgumentException e) {
      System.out.println("Disabling audio: Sound card does not support specified output format");
      return null;
    } catch (LineUnavailableException e) {
      System.out.println("Disabling audio: Sound card busy");
      return null;
    } catch (SecurityException e) {
      System.out.println("Disabling audio: Sound card inaccessible due to security restrictions");
      return null;
    }
  }

  //
  // The audio output thread.
  //
  protected class outputClass implements Runnable {
    public void run() {
      boolean running = false;
      byte[] data;
      int written;

      localSoundSystem.start();
      while (true) {
        synchronized (outputBuffer) {
          // Allow the buffer to fill before beginning playback.
          if (outputThread != null && !running) {
            if (outputBufferSize < outputBufferTargetSize) {
              try { outputBuffer.wait(); } catch (InterruptedException ie) {}
              continue;
            }
            running = true;
          }

          // Retrieve some audio data.
          try {
            data = (byte[])(outputBuffer.removeFirst());
            outputBufferSize -= data.length;
            outputBuffer.notifyAll();
          } catch (NoSuchElementException nsee) {
            // Time to quit.
            if (outputThread == null) { break; }
            // Buffer underflow!  Stop playback until buffer refills.
            //System.out.println("Audio buffer underflow");
            running = false;
            continue;
          }
        }

        // Play the audio data.
        written = localSoundSystem.write(data, 0, data.length);
        statNumBytesOut += written;
        if (written != data.length) {
          System.out.println("Warning: Only wrote " + written + " of " +
            data.length + "bytes of audio data");
        }
      }
      localSoundSystem.drain();
      localSoundSystem.stop();
      localSoundSystem.close();
      localSoundSystem = null;
      outputBuffer = null;
    }
  }

}
